using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Imaging;

namespace Game
{
	public enum Turn_Type
	{
		Turn_Left,
		Move_Fwd,
		Turn_Right,
		Fire_Ammo,
		Fire_Missile,
		Pass,

		None
	};

	/// <summary>
	/// Summary description for Game_Screen.
	/// </summary>
	public class Game_Screen : Screen
	{
		private static int Explosion_Speed = 2;

		private enum Screen_State
		{
			Intro,
			Game,
			Ask_For_Exit,
			Finished,

			Game_Anim_Plr_Move,
			Game_Anim_Plr_Shoot,
			Game_Anim_Plr_Die,
		}

		private Resources resources;

		private Screen_State scr_state;
		private Game_State state;

		private ArrayList turn_buttons = new ArrayList();
		private ArrayList battlelog = new ArrayList();

		private Gfx_Button pushed_button = null;

		private New_Screen_State ret_state = New_Screen_State.None;

		private float ang = 0.0f;

		private ArrayList ai_buttons = new ArrayList();
		private int ai_pause;

		private int moveanim_plr = -1;
		private int moveanim_dx;
		private int moveanim_dy;

		private Player_State shootanim_src;
		private Player_State shootanim_dest;
		private int shootanim_dmg;
		private int shootanim_frame;
		private int shootanim_x;
		private int shootanim_y;
		private int shootanim_dx;
		private int shootanim_dy;
		private int shootanim_left;

		private Player_State dieanim_plr;
		private int dieanim_frame;
		private int dieanim_framecnt;

		private Turn_Type[] moves =
		{
			Turn_Type.Turn_Left, Turn_Type.Turn_Left,
			Turn_Type.Move_Fwd, Turn_Type.Move_Fwd, Turn_Type.Move_Fwd, Turn_Type.Move_Fwd,
			Turn_Type.Turn_Right, Turn_Type.Turn_Right, 
			Turn_Type.Fire_Ammo, Turn_Type.Fire_Ammo, 
			Turn_Type.Fire_Missile,
		};

		public Game_Screen( Resources res, int plrs )
		{
			resources = res;
			scr_state = Screen_State.Intro;
			state = new Game_State( plrs );

			New_Turn();
		}

		#region Screen Members

		public New_Screen_State Update( Mouse_State mouse )
		{
			ang += 0.02f;

			switch ( scr_state )
			{
				case Screen_State.Intro:
					if ( mouse.Is_Released( Mouse_Key.Left ) )
						scr_state = Screen_State.Game;
					break;

				case Screen_State.Game:
					Update_Game( mouse );
					break;

				case Screen_State.Ask_For_Exit:
					if ( mouse.Is_Released( Mouse_Key.Right ) )
						ret_state = New_Screen_State.Menu;
					if ( mouse.Is_Released( Mouse_Key.Left ) )
						scr_state = Screen_State.Game;
					break;

				case Screen_State.Finished:
					if ( mouse.Is_Released( Mouse_Key.Left ) )
						ret_state = New_Screen_State.Menu;
					break;

				case Screen_State.Game_Anim_Plr_Move:
					Update_Move_Anim();
					break;

				case Screen_State.Game_Anim_Plr_Shoot:
					Update_Shoot_Anim();
					break;

				case Screen_State.Game_Anim_Plr_Die:
					Update_Die_Anim();
					break;
			}

			return ret_state;
		}

		// Notifies of a keypress
		public void Key_Pressed( System.Windows.Forms.Keys key )
		{
			if ( key == System.Windows.Forms.Keys.Escape &&
				 scr_state == Screen_State.Game )
			{
				scr_state = Screen_State.Ask_For_Exit;
			}
		}

		private void Update_Game( Mouse_State mouse )
		{
			// Get the mouse location
			Gfx_Button mouse_over = null;
			foreach ( Gfx_Button b in turn_buttons )
			{
				if ( b.Is_In_Button( mouse.x, mouse.y ) )
					mouse_over = b;
			}

			// Handle the players mouse input
			if ( mouse.Is_Pressed( Mouse_Key.Left ) )
				pushed_button = mouse_over;
			if ( !mouse.Is_Down( Mouse_Key.Left) )
				pushed_button = mouse_over;

			if ( mouse.Is_Released( Mouse_Key.Left ) )
			{
				if ( pushed_button == mouse_over &&
					pushed_button != null &&
					!state.players[ state.turn ].is_computer )
				{
					Do_Plr_Action( pushed_button );
				}
				pushed_button = null;
			}

			// Feedback the AI moves?
			if ( state.players[ state.turn ].is_computer )
			{
				--ai_pause;
				if ( ai_pause < 0 )
				{
					Gfx_Button button = (Gfx_Button) ai_buttons[ 0 ];
					ai_buttons.RemoveAt( 0 );

					if ( Do_Plr_Action( button ) )
						ai_pause = 30;
				}
			}
		}

		private void Anim_Over()
		{
			scr_state = Screen_State.Game;

			// Init the next turn (this or next player)
			if ( turn_buttons.Count <= 1 ||
				 state.Count_Alive() <= 1 )
			{
				New_Turn();
			}
		}

		private void Update_Move_Anim()
		{
			int x_amt = 0;
			int y_amt = 0;

			if ( moveanim_dx != 0 ) x_amt = Math.Sign( moveanim_dx ) * 2;
			if ( moveanim_dy != 0 ) y_amt = Math.Sign( moveanim_dy ) * 2;

			moveanim_dx -= x_amt;
			moveanim_dy -= y_amt;

			if ( moveanim_dx == 0 && moveanim_dy == 0 )
			{
				moveanim_plr = -1;
				Anim_Over();
			}
		}

		private void Update_Shoot_Anim()
		{
			int speed = 8;

			shootanim_x += shootanim_dx * speed;
			shootanim_y += shootanim_dy * speed;
			shootanim_left -= speed;

			if ( shootanim_left <= 0 )
			{
				bool does_die = false;

				// Hit the player?
				if ( shootanim_dest != null )
				{
					shootanim_dest.health -= Math.Max( shootanim_dmg, 0 );

					String word = ( shootanim_dmg == 1 ) ? " pt" : " pts";
					Add_Log( shootanim_src.name + " hits " + shootanim_dest.name );
					Add_Log( "  for " + shootanim_dmg + word );

					if ( shootanim_dest.health <= 0 )
					{
						Add_Log( shootanim_dest.name + " dies!" );

						shootanim_src.ammo = Math.Min( shootanim_src.ammo + 1, Player_State.Max_Ammo );
						shootanim_src.health = Math.Min( shootanim_src.health + 1, Player_State.Max_Health );

						dieanim_plr = shootanim_dest;
						dieanim_frame = resources.explosion.Length - 1;
						dieanim_framecnt = Explosion_Speed;

						does_die = true;
					}
				}

				// Next state
				if ( does_die )
					scr_state = Screen_State.Game_Anim_Plr_Die;
				else
					Anim_Over();
			}
		}

		private void Update_Die_Anim()
		{
			--dieanim_framecnt;
			if ( dieanim_framecnt <= 0 )
			{
				dieanim_framecnt = Explosion_Speed;

				--dieanim_frame;
				if ( dieanim_frame < 0 )
				{
					Anim_Over();
				}
			}
		}

		public void Render(System.Drawing.Graphics g, int scr_w, int scr_h)
		{
			int side_width = 160;
			int side_height = scr_h;
			int bottom_width = scr_w - side_width;
			int bottom_height = 64;

			// Render the lower area
			if ( true )
			{
				g.DrawImage( resources.game_larea, new Rectangle( 0, scr_h - bottom_height, bottom_width, bottom_height ), 0, 0, bottom_width, bottom_height, GraphicsUnit.Pixel );

				Brush larea_font_brush = new SolidBrush( Color.FromArgb( 130, 0, 0, 0 ) );
				Font f = resources.font_comic_big;
				Resources.Write_Center( g, larea_font_brush, f, state.players[ state.turn ].name, bottom_width / 2, scr_h - bottom_height / 2 );
			}

			// Render the turn buttons
			foreach ( Gfx_Button b in turn_buttons )
			{
				b.Draw( g, b == pushed_button );
			}

			// Render the side area
			if ( true )
			{
				g.DrawImage( resources.game_side, new Rectangle( scr_w - side_width, 0, side_width, side_height ), 0, 0, side_width, side_height, GraphicsUnit.Pixel );
			}

			// Render the player info
			for ( int i = 0; i < state.players.Length; ++i )
			{
				int x = scr_w - side_width;
				int cx = x + side_width / 2;
				int y = i * 80 + 12;

				Player_State plr = state.players[ i ];

				Resources.Write_Center_Shade(
					g, resources.font_comic_small,
					plr.name, cx, y );

				int img_w = 27;
				int wx = x + 8 + ( plr.ammo - 1 ) * img_w;
				int wy = y + 10;

				for ( int j = 0; j < plr.ammo; ++j )
				{
					Resources.Draw_Image( g, resources.board_items[ 0 ], wx, wy );
					wx -= img_w;
				}

				wx = x + 8 + ( plr.missiles - 1 ) * img_w;
				wy += 26;
				for ( int j = 0; j < plr.missiles; ++j )
				{
					Resources.Draw_Image( g, resources.board_items[ 1 ], wx, wy );
					wx -= img_w;
				}

				wx = x + side_width - 8 - img_w;
				for ( int j = 0; j < plr.health; ++j )
				{
					Resources.Draw_Image( g, resources.icon_heart, wx, wy );
					wx -= 6;
				}

				// Shade the area
				if ( i != state.turn )
				{
					Brush shade = new SolidBrush( Color.FromArgb( 160, 0, 0, 0 ) );
					g.FillRectangle( shade, x + 3, y - 8, side_width - 8, 74 );
				}
			}

			// Render the rules / goals
			if ( true )
			{
				int x = scr_w - side_width;
				int cx = x + side_width / 2;
				int y = 4 * 80 + 12;

				Resources.Write_Center_Shade(
					g, resources.font_comic_small,
					"Battle log", x + side_width / 2, y );

				y += 6;
				foreach ( String s in battlelog )
				{
					Resources.Write_Shade(
						g, resources.font_comic_tiny, 
						s, x + 5, y );
					y += 12;
				}
			}

			// Render the field
			if ( true )
			{
				for ( int y = 0; y < Game_State.Board_Height; ++y )
				{
					for ( int x = 0; x < Game_State.Board_Width; ++x )
					{
						int src_x = (int) state.board[ x, y ] * 32;
						int sx = x * 32;
						int sy = y * 32;

						g.DrawImage( resources.board_tiles, new Rectangle( sx, sy, 32, 32 ), src_x, 0, 32, 32, GraphicsUnit.Pixel );

						if ( state.pieces[ x, y ] != Board_Piece.None )
						{
							Bitmap item = resources.board_items[ (int) state.pieces[ x, y ] ];

							double a = ang + x * 3 + y * 1673;
							int hx = (int) ( Math.Sin( a * 1.64 ) * Math.Cos( a * 0.94 ) * 5 );
							int hy = (int) ( Math.Cos( a * 1.12 ) * Math.Sin( a * 0.62 ) * 5 );

							Resources.Draw_Image( g, item, sx + hx, sy + hy );
						}
					}
				}
			}

			// Render the players
			for ( int i = 0; i < state.players.Length; ++i )
			{
				Player_State plr = state.players[ i ];
				int x = plr.x * 32;
				int y = plr.y * 32;

				if ( plr.health <= 0 )
					continue;

				if ( i == moveanim_plr )
				{
					x += moveanim_dx;
					y += moveanim_dy;
				}

				Resources.Draw_Image( g, resources.player_sprites[ plr.dir ], x, y );

				int cx = x + 16 - Player_State.dir_x[ plr.dir ] * 12;
				int cy = y + 16 - Player_State.dir_y[ plr.dir ] * 12;

				Brush b = new SolidBrush( Color.FromArgb( 90, 0, 0, 0 ) );
				g.FillEllipse( b, cx - 8, cy - 10, 16, 20 );
				Resources.Write_Center_Shade( g, resources.font_comic_small, plr.short_name, cx, cy );
			}

			// Render the foreground
			switch ( scr_state )
			{
				case Screen_State.Intro:
					Render_Intro( g, scr_w, scr_h );
					break;

				case Screen_State.Ask_For_Exit:
					Render_Ask_For_Exit( g, scr_w, scr_h );
					break;

				case Screen_State.Finished:
					Render_Finished( g, scr_w, scr_h );
					break;

				case Screen_State.Game_Anim_Plr_Shoot:
					Resources.Draw_Image_Center( g, resources.bullets[ shootanim_frame ], shootanim_x, shootanim_y );
					break;

				case Screen_State.Game_Anim_Plr_Die:
					if ( true )
					{
						int x = dieanim_plr.x * 32 + 16;
						int y = dieanim_plr.y * 32 + 16;

						Resources.Draw_Image_Center( g, resources.explosion[ dieanim_frame ], x, y );
					}
					break;
			}
		}

		#endregion

		private String intro_text =
			"The objective is to kill the other players before\n" + 
			"they kill you. On each turn you get 3-8 different\n" +
			"move icons. You may play as many of them as you like.\n" +
			"Unused icons are discarded and the turn is over when\n" +
			"you either click 'PASS' or run out of icons. You cannot\n" +
			"move over a gap but you still can shoot over it.\n" +
			"\n" +
			"The possible icons you can have are:\n" +
			" - Turn left\n" +
			" - Turn right\n" +
			" - Move forward\n" +
			" - Shoot\n" +
			" - Fire missile\n" +
			"\n" +
			"On the battleground you have more ammo, missiles and\n" +
			"health. You can have at most 5 ammo, 3 missiles and 8\n" +
			"health. Each ammo make 1 point of damage and each\n" +
			"missile hits for 3 points. The killer is rewarded with\n" +
			"one extra ammo and health.";

		private void Render_Intro(System.Drawing.Graphics g, int scr_w, int scr_h)
		{
			int w = (int) ( scr_w * 0.7f );
			int h = (int) ( scr_h * 0.8f );
			int ux = ( scr_w - w ) / 2;
			int uy = ( scr_h - h ) / 2;
			int cx = ux + w / 2;

			int line_h = 16;

			Brush b = new SolidBrush( Color.FromArgb( 210, 0, 0, 0 ) );
			g.FillRectangle( b, ux, uy, w, h );

			Font f = resources.font_comic_small;

			int x = ux + w / 64;
			int y = uy + 12;
			Resources.Write_Center_Shade( g, f, "Welcome to Radox!", cx, y ); y += line_h;

			String[] lines = intro_text.Split( '\n' );
			foreach ( String s in lines )
			{
				Resources.Write_Shade( g, f, s, x, y );
				y += line_h;
			}

			y = uy + h - line_h;
			Resources.Write_Center_Shade( g, f, "Press left mouse button to continue!", cx, y );
		}

		private void Render_Ask_For_Exit(System.Drawing.Graphics g, int scr_w, int scr_h)
		{
			int w = (int) ( scr_w * 0.5f );
			int h = (int) ( scr_h * 0.17f );
			int ux = ( scr_w - w ) / 2;
			int uy = ( scr_h - h ) / 2;
			int cx = ux + w / 2;

			Brush b = new SolidBrush( Color.FromArgb( 170, 0, 0, 0 ) );
			g.FillRectangle( b, ux, uy, w, h );

			int y = uy + 20;
			Resources.Write_Center_Shade( g, resources.font_comic_medium, "Exit to main menu?", cx, y );

			y = uy + h - 32;
			Resources.Write_Center_Shade( g, resources.font_comic_small, "Press left mouse button to continue or", cx, y );
			y += 16;
			Resources.Write_Center_Shade( g, resources.font_comic_small, "right mouse button to exit!", cx, y );
		}

		private void Render_Finished(System.Drawing.Graphics g, int scr_w, int scr_h)
		{
			int w = (int) ( scr_w * 0.5f );
			int h = (int) ( scr_h * 0.17f );
			int ux = ( scr_w - w ) / 2;
			int uy = ( scr_h - h ) / 2;
			int cx = ux + w / 2;

			String msg = state.players[ state.turn ].name + " won!";

			Brush b = new SolidBrush( Color.FromArgb( 170, 0, 0, 0 ) );
			//g.FillRectangle( b, ux, uy, w, h );
			g.FillRectangle( b, 0, 0, scr_w, scr_h );

			int y = uy + 20;
			Resources.Write_Center_Shade( g, resources.font_comic_medium, msg.ToUpper(), cx, y );

			y = uy + h - 16;
			Resources.Write_Center_Shade( g, resources.font_comic_small, "Press left mouse button to continue!", cx, y );
		}

		private bool Do_Plr_Action( Gfx_Button button )
		{
			Player_State plr = state.players[ state.turn ];

			bool ok = true;

			Turn_Type turn = (Turn_Type) button.user_data;
			switch ( turn )
			{
				case Turn_Type.Turn_Left:
					plr.dir = ( plr.dir + 4 - 1 ) % 4;
					break;

				case Turn_Type.Turn_Right:
					plr.dir = ( plr.dir + 4 + 1 ) % 4;
					break;

				case Turn_Type.Move_Fwd:
					ok = state.Is_Valid_Dir( plr, Player_State.dir_x[ plr.dir ], Player_State.dir_y[ plr.dir ] );
					if ( ok )
					{
						int dx = Player_State.dir_x[ plr.dir ];
						int dy = Player_State.dir_y[ plr.dir ];

						// Move the player
						plr.x += dx;
						plr.y += dy;

						// Animate the movement
						scr_state = Screen_State.Game_Anim_Plr_Move;
						moveanim_plr = state.turn;
						moveanim_dx = -dx * 32;
						moveanim_dy = -dy * 32;

						// Collected something?
						Board_Piece piece = state.pieces[ plr.x, plr.y ];
						switch ( state.pieces[ plr.x, plr.y ] )
						{
							case Board_Piece.Ammo:
								plr.ammo = Math.Min( plr.ammo + 1, Player_State.Max_Ammo );
								break;

							case Board_Piece.Missile:
								plr.missiles = Math.Min( plr.missiles + 1, Player_State.Max_Missiles );
								break;

							case Board_Piece.Health:
								plr.health = Math.Min( plr.health + 1, Player_State.Max_Missiles );
								break;
						}

						if ( piece != Board_Piece.None )
						{
							state.pieces[ plr.x, plr.y ] = Board_Piece.None;
							state.Add_Random_Piece( piece );
						}
					}
					break;

				case Turn_Type.Fire_Ammo:
					if ( plr.ammo <= 0 )
						ok = false;
					else
					{
						plr.ammo--;
						Shoot( plr, 1 );
					}
					break;

				case Turn_Type.Fire_Missile:
					if ( plr.missiles <= 0 )
						ok = false;
					else
					{
						plr.missiles--;
						Shoot( plr, 3 );
					}
					break;

				case Turn_Type.Pass:
					New_Turn();
					return true;

				default:
					ok = false;
					break;
			}

			if ( !ok )
				return false;

			// Remove the pushed button
			turn_buttons.Remove( button );

			// Simulate "animation over" for the turnings
			if ( turn == Turn_Type.Turn_Left ||
				turn == Turn_Type.Turn_Right )
				Anim_Over();

			return true;
		}

		private Player_State Cast_Ray( int x, int y, int dx, int dy, out int amt )
		{
			amt = 0;

			while ( true )
			{
				x += dx;
				y += dy;

				if ( x < 0 || y < 0 || x >= Game_State.Board_Width || y >= Game_State.Board_Height )
					break;
				if ( state.board[ x, y ] == Board_Block.Wall )
					break;

				foreach ( Player_State p in state.players )
				{
					if ( p.health > 0 && p.x == x && p.y == y )
						return p;
				}

				++amt;
			}

			return null;
		}

		private void Shoot( Player_State plr, int dmg )
		{
			int dx = Player_State.dir_x[ plr.dir ];
			int dy = Player_State.dir_y[ plr.dir ];
			int amt;

			Player_State hit = Cast_Ray( plr.x, plr.y, dx, dy, out amt );

			shootanim_src = plr;
			shootanim_dest = hit;
			shootanim_dmg = dmg;
			shootanim_frame = ( dmg == 1 ) ? 0 : 1;
			shootanim_x = plr.x * 32 + 16;
			shootanim_y = plr.y * 32 + 16;
			shootanim_dx = dx;
			shootanim_dy = dy;
			shootanim_left = amt * 32 + 16;

			scr_state = Screen_State.Game_Anim_Plr_Shoot;

			if ( hit != null )
			{
				shootanim_left += 16;
			}
		}

		private void New_Turn()
		{
			turn_buttons.Clear();

			// Game over?
			if ( state.Count_Alive() <= 1 )
			{ // GAME OVER!
				scr_state = Screen_State.Finished;
				return;
			}

			// Get the next turn
			do
			{
				state.turn = ( state.turn + 1 ) % state.players.Length;
			} while ( state.players[ state.turn ].health <= 0 );

			// Init the turn
			Player_State plr = state.players[ state.turn ];

			int amt = Resources.random.Next( 3, 8 );

			for ( int i = 0; i < amt; ++i )
			{
				Turn_Type but = moves[ Resources.random.Next( moves.Length ) ];
				Add_Turn_Button( resources.turn_icons[ (int) but ], but );
			}
			Add_Turn_Button( resources.turn_icons[ (int) Turn_Type.Pass ], Turn_Type.Pass );

			// Init the AI (if needed)
			if ( plr.is_computer )
				Do_AI( plr );
		}

		private void Add_Turn_Button( Bitmap icon, Turn_Type t )
		{
			int w = 48;
			int h = 48;

			int x = 8 + turn_buttons.Count * ( w + 8 );
			int y = 480 - 64 + 8;

			Gfx_Button button = new Gfx_Button( icon, new Rectangle( x, y, w, h ), false );
			button.user_data = (int) t;
			turn_buttons.Add( button );
		}

		private void Add_Log( String msg )
		{
			battlelog.Add( msg );
			while ( battlelog.Count > 10 )
			{
				battlelog.RemoveAt( 0 );
			}
		}

		#region AI code

		private void Do_AI( Player_State comp )
		{
			Player_State human = state.players[ 0 ];

			// Init
			ai_buttons.Clear();
			ai_pause = 15;

			// Get the goal point
			int goal_x = human.x;
			int goal_y = human.y;

			if ( comp.ammo <= 0 && comp.missiles <= 0 )
			{ // Find the nearest ammo/missile
				int cost = 999999;

				for ( int y = 0; y < Game_State.Board_Height; ++y )
				{
					for ( int x = 0; x < Game_State.Board_Height; ++x )
					{
						if ( state.pieces[ x, y ] == Board_Piece.None )
							continue;

						int dx = x - comp.x;
						int dy = y - comp.y;
						int dist = ( dx * dx ) + ( dy * dy );

						if ( dist < cost )
						{
							cost = dist;
							goal_x = x;
							goal_y = y;
						}
					}
				}
			}

			// Collect the turns
			Turn_Type[] turns = new Turn_Type[ turn_buttons.Count - 1 ];
			int left_amt = 0;
			int right_amt = 0;
			int fwd_amt = 0;

			for ( int i = 0; i < turns.Length; ++i )
			{
				turns[ i ] = Get_Turn_Button_Type( i );

				switch ( turns[ i ] )
				{
					case Turn_Type.Turn_Left:		left_amt++; break;
					case Turn_Type.Turn_Right:		right_amt++; break;
					case Turn_Type.Move_Fwd:		fwd_amt++; break;
				}
			}

			// Try to turn towards it
			int goal_dx = goal_x - comp.x;
			int goal_dy = goal_y - comp.y;

			if ( Math.Abs( goal_dx ) > Math.Abs( goal_dy ) )
				goal_dy = 0;
			else
				goal_dx = 0;

			int goal_dfwd = Math.Max( Math.Abs( goal_dx ), Math.Abs( goal_dy ) );

			if ( goal_dx != 0 ) goal_dx = Math.Sign( goal_dx );
			if ( goal_dy != 0 ) goal_dy = Math.Sign( goal_dy );

			int req_right = 0;

			for ( int i = 0; i < 4; ++i )
			{
				int ndir = ( comp.dir + i ) % 4;
				if ( goal_dx == Player_State.dir_x[ ndir ] &&
					 goal_dy == Player_State.dir_y[ ndir ] )
				{
					break;
				}

				++req_right;
			}

			int req_left = 0;
			if ( req_right > 0 )
				req_left = 4 - req_right;

			// One or the other satisfies?
			if ( req_left > 0 && req_right > 0 )
			{
				if ( left_amt >= req_left )
				{ // Left is enough!
					for ( int i = 0; i < turns.Length; ++i )
					{
						if ( turns[ i ] == Turn_Type.Turn_Left )
						{
							turns[ i ] = Turn_Type.None;
							ai_buttons.Add( turn_buttons[ i ] );
							--req_left;
							if ( req_left <= 0 )
								break;
						}
					}

					req_left = req_right = 0;
				}
				else if ( right_amt >= req_right )
				{ // Right is enough!
					for ( int i = 0; i < turns.Length; ++i )
					{
						if ( turns[ i ] == Turn_Type.Turn_Right )
						{
							turns[ i ] = Turn_Type.None;
							ai_buttons.Add( turn_buttons[ i ] );
							--req_right;
							if ( req_right <= 0 )
								break;
						}
					}

					req_left = req_right = 0;
				}
			}

			// Try to move forward (if not turning)
			if ( req_left == 0 && goal_dfwd > 0 )
			{
				for ( int i = 0; i < turns.Length; ++i )
				{
					if ( turns[ i ] == Turn_Type.Move_Fwd )
					{
						turns[ i ] = Turn_Type.None;
						ai_buttons.Add( turn_buttons[ i ] );
						--goal_dfwd;
						if ( goal_dfwd <= 0 )
							break;
					}
				}
			}

			// Shoot if line of sight
			if ( req_left == 0 )
			{
				int amt;

				if ( Cast_Ray( comp.x, comp.y, goal_dx, goal_dy, out amt ) != null )
				{
					for ( int i = 0; i < turns.Length; ++i )
					{
						if ( turns[ i ] == Turn_Type.Fire_Ammo ||
							 turns[ i ] == Turn_Type.Fire_Missile )
						{
							turns[ i ] = Turn_Type.None;
							ai_buttons.Add( turn_buttons[ i ] );
						}
					}
				}
			}

			// Done! (add the final "pass" just in case)
			ai_buttons.Add( turn_buttons[ turn_buttons.Count - 1 ] );
		}

		private Turn_Type Get_Turn_Button_Type( int idx )
		{
			return (Turn_Type) ( ((Gfx_Button)turn_buttons[ idx]).user_data );
		}

		#endregion
	}
}
