using System;
using System.Collections;

namespace Game
{
	public enum Board_Block
	{
		Wall,
		Hole,
		Normal,

		__COUNT__
	}

	public enum Board_Piece
	{
		Ammo,
		Missile,
		Health,
		None,

		__COUNT__
	}

	public class Player_State
	{
		public static int Max_Health = 8;
		public static int Max_Ammo = 5;
		public static int Max_Missiles = 3;

		public static int[] dir_x = { -1, 0, 1, 0 };
		public static int[] dir_y = { 0, -1, 0, 1 };

		public int idx;

		public String name;
		public String short_name;
		public bool is_computer;

		public int x, y;
		public int dir;

		public int health;

		public int ammo;
		public int missiles;

		public Player_State( int idx, String name )
		{
			this.idx = idx;
			this.name = name;
			short_name = "";
			is_computer = false;

			dir = Resources.random.Next( 4 );
			health = 7;
			ammo = 3;
			missiles = 1;
		}
	}

	/// <summary>
	/// Summary description for Game_State.
	/// </summary>
	public class Game_State
	{
		public static int Board_Width = 15;
		public static int Board_Height = 13;

		public Player_State[] players;

		public Board_Block[,] board = new Board_Block[ Board_Width, Board_Height ];
		public Board_Piece[,] pieces = new Board_Piece[ Board_Width, Board_Height ];

		public int turn;

		private String[] player_names = { "Player 1", "Player 2", "Player 3", "Player 4" };

		public Game_State( int plr_cnt )
		{
			int block_percent = 20;

			if ( plr_cnt == 1 )
				block_percent = 10;

			// Init the board
			for ( int y = 0; y < Board_Height; ++y )
			{
				for ( int x = 0; x < Board_Width; ++x )
				{
					board[ x, y ] = Board_Block.Normal;
					pieces[ x, y ] = Board_Piece.None;

					if ( Resources.random.Next( 100 ) < block_percent )
						board[ x, y ] = (Board_Block) ( Resources.random.Next( 2 ) );
				}
			}

			// Init the players
			int cnt = Math.Max( plr_cnt, 2 );
			players = new Player_State[ cnt ];

			for ( int i = 0; i < players.Length; ++i )
			{
				players[ i ] = new Player_State( i, player_names[ i ] );
				players[ i ].short_name = ( i + 1 ).ToString();

				// Pick the starting position
				bool ok = false;
				while ( !ok )
				{
					int x = Resources.random.Next( Board_Width );
					int y = Resources.random.Next( Board_Height );

					if ( board[ x, y ] != Board_Block.Normal )
						continue;

					ok = true;
					foreach ( Player_State plr in players )
					{
						if ( plr != null &&
							( plr.x == x || plr.y == y ) )
						{
							ok = false;
							break;
						}
					}

					if ( ok )
					{
						players[ i ].x = x;
						players[ i ].y = y;
					}
				}
			}

			if ( plr_cnt == 1 )
			{
				players[ 0 ].name = "Human";
				players[ 0 ].short_name = "H";

				players[ 1 ].name = "Computer";
				players[ 1 ].short_name = "C";
				players[ 1 ].is_computer = true;
			}

			// Add the items
			for ( int i = 0; i < cnt * 2; ++i )
			{
				Add_Random_Piece( Board_Piece.Ammo );
			}
			for ( int i = 0; i < cnt; ++i )
			{
				Add_Random_Piece( Board_Piece.Missile );
			}
			Add_Random_Piece( Board_Piece.Health );
			if ( cnt >= 4 )
				Add_Random_Piece( Board_Piece.Health );

			// Init the rest
			turn = Resources.random.Next( players.Length );
		}

		public int Count_Alive()
		{
			int alive = 0;
			foreach ( Player_State plr in players )
			{
				if ( plr.health > 0 )
					alive++;
			}
			return alive;
		}

		public bool Is_Valid_Dir( Player_State plr, int dx, int dy )
		{
			return Is_Valid_Dir( plr.x, plr.y, dx, dy );
		}

		public bool Is_Valid_Dir( int fx, int fy, int dx, int dy )
		{
			int nx = fx + dx;
			int ny = fy + dy;

			// Crossing border?
			if ( nx < 0 || nx >= Board_Width || ny < 0 || ny >= Board_Height )
			{
				return false;
			}

			// Entering another players space?
			foreach ( Player_State p in players )
			{
				if ( p.health > 0 && p.x == nx && p.y == ny )
					return false;
			}

			// Forbidden square?
			switch ( board[ nx, ny ] )
			{
				case Board_Block.Hole:
				case Board_Block.Wall:
					return false;
			}

			// Valid!
			return true;
		}


		public void Add_Random_Piece( Board_Piece type )
		{
			int x;
			int y;

			while ( true )
			{
				x = Resources.random.Next( Board_Width );
				y = Resources.random.Next( Board_Height );

				if ( board[ x, y ] != Board_Block.Normal )
					continue;
				if ( pieces[ x, y ] != Board_Piece.None )
					continue;

				bool is_plr = false;
				foreach ( Player_State plr in players )
				{
					if ( plr.health > 0 && plr.x == x && plr.y == y )
						is_plr = true;
				}
				if ( is_plr )
					continue;

				// Valid!
				pieces[ x, y ] = type;
				return;
			}
		}
	}
}
